/*
** Module   : NativeInteger.java
** Abstract : descriptor for integer data as an argument or return value to a native function 
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20131212 First version.
** 002 GES 20140212 Added a flag to enable different rendering between scalar and array elements.
**                  Added a method to properly emit sign-extended values with word-size width.
*/

package com.goldencode.p2j.library;

import java.io.*;
import com.goldencode.p2j.util.*;

/**
 * A descriptor for integer data as an argument or return value to a native function.
 */
public abstract class NativeInteger
extends NativePrimitive
{
   /** The stored value. */
   protected long value;
   
   /** Flag to allow differential rendering between scalar and array elements. */
   private boolean array = false; 
   
   /**
    * Default constructor.
    */
   public NativeInteger()
   {
      super();
   }
   
   /**
    * Access the data element.
    *
    * @return   The integer data element.
    */
   public long getValue()
   {
      return value;
   }
   
   /**
    * Assign the data element.
    *
    * @param    value
    *           The new integer data element.
    */
   public void setValue(long value)
   {
      this.value = value;
   }
   
   /**
    * Reports if this data element is an array member.
    *
    * @return   <code>true</code> if this is an element of an array.
    */
   public boolean isInArray()
   {
      return array;
   }
   
   /**
    * Defines whether or not the data element is an array member.
    *
    * @param    array
    *           <code>true</code> if this is an element of an array.
    */
   public void setInArray(boolean array)
   {
      this.array = array;
   }
   
   /**
    * This method should be used to render signed scalar elements such that the written data is
    * sign-extended to the word-size.  This is how Progress 4GL handles integer data. The reason
    * this is needed is because scalar values are passed on the stack in a "slot" that is the
    * width of the platform's word-size. Even if the native code is expecting something smaller,
    * the actual memory passed will have a larger, sign-extended form of the data.
    * <p>
    * This doesn't apply to non-integer data.  It also doesn't apply to arrays, since there is
    * no word-size "padding" in array elements. Each native array's element is exactly the size
    * of the element itself and so there is no sign-extending to do.
    *
    * @param    ptr
    *           The buffer in which to write at position 1.
    */
   public void writeNativeWordSize(memptr ptr)
   {
      if (LibraryManager.ptrSize() == 4)
      {
         ptr.setLong(value, 1);
      }
      else
      {
         ptr.setInt64(value, 1);
      }
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      
      // don't assign the member directly, subclasses may have some behavior hooked to this
      setValue(in.readLong());
      
      array = in.readBoolean();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeLong(value);
      out.writeBoolean(array);
   }
}
