/*
** Module   : NativeString.java
** Abstract : descriptor for text data as an argument or return value to a native function 
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20140107 First version.
** 002 GES 20140201 Complete rewrite of the render() method to override the parent's version. It
**                  is important to write strings into their own memory buffer and then write the
**                  pointer into the given buffer. On restore(), the parent's version is fine.
*/

package com.goldencode.p2j.library;

import java.io.*;
import java.math.*;

import com.goldencode.p2j.util.*;

/**
 * A descriptor for text data as an argument or return value to a native function.
 */
public class NativeString
extends NativePrimitive
{
   /** Amount of extra bytes allocated to reduce the change of buffer overflows. */
   private static int BUFFER_PADDING = 128;
   
   /** The stored value (defaults to empty string). */
   protected String value = ""; 
   
   /**
    * Default constructor.
    */
   public NativeString()
   {
      super();
      type = LibraryManager.TYPE_POINTER;
   }
   
   /**
    * Access the data element.
    *
    * @return   The doubleing point data element.
    */
   public String getValue()
   {
      return value;
   }
   
   /**
    * Assign the data element.
    *
    * @param    value
    *           The new text data element.  Ignored if <code>null</code>.
    */
   public void setValue(String value)
   {
      if (value != null)
         this.value = value;
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      value = (String) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeObject(value);
   }
   
   /**
    * Report the size (in bytes) of the native memory needed to store this instance.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @return   This is always the size of a pointer on the client platform.
    */
   public int size()
   {
      return LibraryManager.ptrSize();
   }
   
   /**
    * Copy our data into a large-enough buffer in native memory and then render the buffer's
    * address into native memory if an explicit address is provided and return the data's
    * address to the caller.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address into which the address should be rendered, if the caller
    *           is supplying a buffer or 0 if the internally created buffer's address should be
    *           returned directly.
    *
    * @return   The memory address of the buffer or 0 if there is no data to output.
    */
   public long render(long addr)
   {
      // make sure there is room for the null terminator and the value itself
      int sz = value.length() + 1 + BUFFER_PADDING;
      
      // save off the buffer for restore later
      ref = new memptr();
      ref.setLength(sz);
      
      // since we are always passed as a pointer, we always copy our data into the
      // native buffer; this will always result in at least the empty string (a null
      // terminator in the first position of the buffer)
      writeNative(ref);
      
      // - return parms (always scalar) do not call through here
      // - non-return scalar parm types will have addr == 0 and we just need to return the
      //   target address of our newly created buffer memptr
      // - array elements will be called with addr != 0 because the address needs to be
      //   written to memory controlled by the caller
      
      // return our buffer's address
      long target = ref.getPointerValue().longValue();
      
      // write it into the caller's buffer if specified
      if (addr != 0)
      {
         memptr local = new memptr();
         local.setPointerValue(addr);
         
         if (size() == 4)
         {
            local.setLong(target, 1);
         }
         else
         {
            local.setInt64(target, 1);
         }
      }
      
      return target;
   }
   
   /**
    * Worker method to read the primitive value from the first index position of the given
    * memory buffer.
    *
    * @param    ptr
    *           The memory buffer to reference.
    */
   protected void readNative(memptr ptr)
   {
      character txt = ptr.getString(1);
      
      if (!txt.isUnknown())
      {
         value = txt.toStringMessage();
      }
   }
   
   /**
    * Worker method to write the primitive value to the first index position of the given
    * memory buffer.
    *
    * @param    ptr
    *           The memory buffer to reference.
    */
   protected void writeNative(memptr ptr)
   {
      ptr.setString(value, 1);
   }
}
