/*
** Module   : NativeTypeArray.java
** Abstract : an array of descriptors passed as a single argument to a native function 
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20131211 First version.
** 002 GES 20140130 Even arrays need to be passed as a pointer to a pointer (one more level of
**                  indirection than we were providing).
*/

package com.goldencode.p2j.library;

import java.io.*;
import com.goldencode.p2j.util.*;

/**
 * An array of descriptors passed as a single argument to a native function.  4GL RETURN
 * parameters cannot have an extent, so this can only be used for INPUT, INPUT-OUTPUT and
 * OUTPUT parameters.  This class hides the complexity of rendering/restoring arrays to
 * native memory.
 */
public class NativeTypeArray
extends BaseNativeType
{
   /** Stores the contained elements for the array. */
   private BaseNativeType[] elements = new BaseNativeType[0];
   
   /**
    * Default constructor.
    */
   public NativeTypeArray()
   {
      super();
      setPassByPointer(true);
      type = LibraryManager.TYPE_POINTER;
   }
   
   /**
    * Constructor that takes a list of elements.
    *
    * @param    elements
    *           The elements for this array.
    */
   public NativeTypeArray(BaseNativeType[] elements)
   {
      this();
      this.elements = elements;
   }
   
   /**
    * Access the array of data elements.
    *
    * @return   The array of elements.
    */
   public BaseNativeType[] getElements()
   {
      return elements;
   }
   
   /**
    * Assign the array of data elements.
    *
    * @param    elements
    *           The array of elements.
    */
   public void setElements(BaseNativeType[] elements)
   {
      this.elements = elements;
   }
   
   /**
    * Report the size (in bytes) of the native memory needed to store this instance.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @return   The number of bytes needed.
    */
   public int size()
   {
      return (elements.length > 0) ? elements.length * elements[0].size() : 0;
   }
   
   /**
    * Render the data into native memory, which will be allocated if the address is not
    * given.
    * <p>
    * This should never be called on the server, only on the client side.
    * <p>
    * Arrays are rendered as a pointer to a block of memory in which each element is
    * written contiguously from lowest address to highest address and the size of the
    * block is the number of elements multiplied by the size of each element (which 
    * must be the same for all elements). This is consistent with how native code ('C'
    * language) handles arrays.
    *
    * @param    addr
    *           The memory address to which the data should be rendered, if the caller
    *           is supplying a buffer or 0 if the buffer should be allocated.
    *
    * @return   The memory address of the rendered data or 0 if there is no data to
    *           output.
    */
   public long render(long addr)
   {
      if (elements.length > 0)
      {
         // all elements of the array must have the same size
         int sz = elements[0].size();
         
         memptr local = new memptr();
         
         if (addr == 0)
         {
            // save off for restore use later
            ref = local;
            
            // allocate the block which must be big enough to store the array
            local.setLength(elements.length * sz);
            
            addr = local.getPointerValue().longValue();
         }
         else
         {
            local.setPointerValue(addr);
         }
         
         if (copyIn || elements[0].isPassByPointer())
         {
            for (int i = 0; i < elements.length; i++)
            {
               // calculate the address at which to write our next element
               long idx = addr + (i * sz);
               
               // give each element a chance to render itself
               elements[i].render(idx);
            }
         }
      }
      else
      {
         addr = 0;
      }
      
      return addr;
   }
   
   /**
    * Copy back (if necessary) data from the native memory and deallocate the memory
    * if this instance allocated it in the first place.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address from which the data should be restored, if the caller
    *           is supplying a buffer or 0 if the buffer should already exist.
    */
   public void restore(long addr)
   {
      memptr local = ref;
      
      if (local == null && addr != 0)
      {
         local = new memptr();
         local.setPointerValue(addr);
      }
      
      if (copyOut && local != null && elements.length > 0)
      {
         addr = local.getPointerValue().longValue();
         
         // all elements of the array must have the same size
         int sz = elements[0].size();
         
         // each element must have a chance to restore and deallocate the native resources
         for (int i = 0; i < elements.length; i++)
         {
            // calculate the address at which to read our next element
            long idx = addr + (i * sz);
            
            elements[i].restore(idx);
         }
      }
      
      // deallocate memory if we allocated it
      if (ref != null)
         ref.setLength(0);
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      
      int len = in.readInt();
      
      elements = new BaseNativeType[len];
      
      for (int i = 0; i < len; i++)
      {
         elements[i] = (BaseNativeType) in.readObject();
      }
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      
      out.writeInt(elements.length);
      
      for (int i = 0; i < elements.length; i++)
      {
         out.writeObject(elements[i]);
      }
   }
}
